/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package fr.loof.fonzie;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.LockModeType;
import javax.persistence.Query;

import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import fr.loof.fonzie.keywords.Aggregate;

/**
 * @author <a href="nicolas@apache.org">Nicolas De Loof</a>
 *
 */
@Aspect
public class ORMAspect extends AbstractORMAspect
{
    protected static Logger logger = LoggerFactory.getLogger( ORMAspect.class );

    @Around( "call( * (@javax.persistence.Entity *).as( Class ) )" )
    public Object as( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        Class<?> type = (Class<?>) jp.getArgs()[0];
        if ( type.isInstance( entity ) )
        {
            return entity;
        }
        return entityManager.find( type, JPAUtils.getId( entity ) );
    }

    @Around( "call( boolean (@javax.persistence.Entity *).instanceOf( Class ) )" )
    public boolean instanceOf( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        Class<?> type = (Class<?>) jp.getArgs()[0];
        if ( entity instanceof org.hibernate.proxy.HibernateProxy )
        {
            entity = ( (org.hibernate.proxy.HibernateProxy) entity ).writeReplace();
        }
        // TODO code quivalent pour d'autres implmentations JPA ?
        return type.isInstance( entity );
    }

    @Around( "call( * (@javax.persistence.Entity *).persist() )" )
    public void persist( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        entityManager.persist( entity );
    }

    @Around( "call( * (@javax.persistence.Entity *).merge() )" )
    public Object merge( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        return entityManager.merge( entity );
    }

    @Around( "call( * (@javax.persistence.Entity *).remove() )" )
    public void remove( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        entityManager.remove( entity );
    }

    @Around( "call( static * (@javax.persistence.Entity *).find(..) )" )
    @SuppressWarnings( "unchecked" )
    public Object findById( ProceedingJoinPoint jp )
    {
        Class entityClass = jp.getSignature().getDeclaringType();
        Object primaryKey = jp.getArgs()[0];
        return entityManager.find( entityClass, primaryKey );
    }

    @Around( "call( static * (@javax.persistence.Entity *).getReference(..) )" )
    @SuppressWarnings( "unchecked" )
    public Object getReference( ProceedingJoinPoint jp )
    {
        Class entityClass = jp.getSignature().getDeclaringType();
        Object primaryKey = jp.getArgs()[0];
        return entityManager.getReference( entityClass, primaryKey );
    }

    @Around( "call( * (@javax.persistence.Entity *).lock( javax.persistence.LockModeType ) )" )
    public void lock( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        LockModeType lockMode = (LockModeType) jp.getArgs()[0];
        entityManager.lock( entity, lockMode );
    }

    @Around( "call( * (@javax.persistence.Entity *).refresh() )" )
    public void refresh( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        entityManager.refresh( entity );
    }

    @Around( "call( boolean (@javax.persistence.Entity *).isManaged() )" )
    public boolean contains( ProceedingJoinPoint jp )
    {
        Object entity = jp.getTarget();
        return entityManager.contains( entity );
    }

    @Around( "call( static * (@javax.persistence.Entity *).findWithCriterias(..) )" )
    @SuppressWarnings( "unchecked" )
    public Object findWithCriterias( ProceedingJoinPoint jp )
    {
        MethodSignature signature = (MethodSignature) jp.getSignature();
        final Object[] args = jp.getArgs();

        final Class entityClass = signature.getDeclaringType();
        final Criteria[] criterias = (Criteria[]) args[0];

        PersistenceQuery query = new PersistenceQuery()
        {
            public Class getEntity()
            {
                return entityClass;
            }

            public Query createQuery( EntityManager entityManager )
            {
                StringBuilder stb = new StringBuilder( "select e from " + entityClass.getSimpleName() + " e" );
                String separator = " where ";
                for ( Criteria criteria : criterias )
                {
                    if ( criteria.isEmpty() )
                    {
                        continue;
                    }
                    stb.append( separator );
                    separator = " and ";
                    stb.append( criteria.toString() );
                }
                return entityManager.createQuery( stb.toString() );
            }

            public Iterator<String> parameters()
            {
                List<String> names = new LinkedList<String>();
                for ( Criteria criteria : criterias )
                {
                    names.addAll( criteria.getParams().keySet() );
                }
                return names.iterator();
            }
        };

        List<Object> values = new ArrayList<Object>();
        for ( Criteria criteria : criterias )
        {
            values.addAll( criteria.getParams().values() );
        }
        logger.debug( "Execute parsed 'findByCriteria' query {}", query );
        return queryExecutor.execute( query, values.toArray(), isSingleResult( signature ) );
    }

    @Around( "call( static * (@javax.persistence.Entity *).findWhere*(..) )" )
    public Object findWhere( ProceedingJoinPoint jp )
    {
        return proceedFind( jp, "findWhere" );
    }

    @Around( "call( static * (@javax.persistence.Entity *).findBy*(..) )" )
    public Object findBy( ProceedingJoinPoint jp )
    {
        return proceedFind( jp, "findBy" );
    }

    @Around( "call( static * (@javax.persistence.Entity *).findAll(..) )" )
    @SuppressWarnings( "unchecked" )
    public Object findAll( ProceedingJoinPoint jp )
    {
        MethodSignature signature = (MethodSignature) jp.getSignature();
        Class[] types = signature.getParameterTypes();
        Object[] args = jp.getArgs();

        Class entityClass = signature.getDeclaringType();
        PersistenceQuery query = queryParser.parse( entityClass, Aggregate.ALL, "", types );
        logger.debug( "Execute parsed 'findAll' query {}", query );
        return queryExecutor.execute( query, args, isSingleResult( signature ) );
    }

    @Around( "call( static long (@javax.persistence.Entity *).countBy*(..) )" )
    @SuppressWarnings( "unchecked" )
    public Object countBy( ProceedingJoinPoint jp )
    {
        MethodSignature signature = (MethodSignature) jp.getSignature();
        Class[] types = signature.getParameterTypes();
        Object[] args = jp.getArgs();

        Class entityClass = signature.getDeclaringType();
        String name = signature.getName();
        String criteria = name.substring( "countBy".length() );
        PersistenceQuery query = queryParser.parse( entityClass, Aggregate.COUNT, criteria, types );
        logger.debug( "Execute parsed 'countBy' query {}", query );
        return queryExecutor.execute( query, args, isSingleResult( signature ) );
    }

    @Around( "call( static long (@javax.persistence.Entity *).count() )" )
    @SuppressWarnings( "unchecked" )
    public Object count( ProceedingJoinPoint jp )
    {
        MethodSignature signature = (MethodSignature) jp.getSignature();
        Class[] types = signature.getParameterTypes();
        Object[] args = jp.getArgs();

        Class entityClass = signature.getDeclaringType();
        PersistenceQuery query = queryParser.parse( entityClass, Aggregate.COUNT, "", types );
        logger.debug( "Execute parsed 'count' query {}", query );
        return queryExecutor.execute( query, args, isSingleResult( signature ) );
    }

    @SuppressWarnings( "unchecked" )
    protected final Object proceedFind( ProceedingJoinPoint jp, String prefix )
    {
        MethodSignature signature = (MethodSignature) jp.getSignature();
        Class[] types = signature.getParameterTypes();
        Object[] args = jp.getArgs();

        Class entityClass = signature.getDeclaringType();
        String name = signature.getName();
        String criteria = name.substring( prefix.length() );
        PersistenceQuery query = queryParser.parse( entityClass, Aggregate.ALL, criteria, types );
        logger.debug( "Execute parsed 'findBy' query {}", query );
        return queryExecutor.execute( query, args, isSingleResult( signature ) );
    }

    protected boolean isSingleResult( MethodSignature method )
    {
        return !Collection.class.isAssignableFrom( method.getReturnType() );
    }

}
